# Generated by Django 2.1.5 on 2019-01-30 17:03

import django.contrib.postgres.fields.jsonb
from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0008_provider_authentication'),
    ]

    operations = [
        migrations.AddField(
            model_name='provider',
            name='agency_api_authentication',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=mds.models.agency_api_authentication_default, verbose_name='API Agency Authentication'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-02-15 09:15

import datetime
from django.db import migrations, models
from django.utils.timezone import utc
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0009_provider_agency_api_authentication'),
    ]

    operations = [
        migrations.AddField(
            model_name='device',
            name='dn_battery_pct',
            field=models.FloatField(null=True),
        ),
        migrations.AlterField(
            model_name='area',
            name='creation_date',
            field=models.DateTimeField(default=datetime.datetime(2012, 1, 1, 0, 0, tzinfo=utc)),
        ),
        migrations.AlterField(
            model_name='area',
            name='deletion_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='area',
            name='label',
            field=mds.models.UnboundedCharField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='eventrecord',
            name='event_type',
            field=mds.models.UnboundedCharField(choices=[('service_start', 'Service start'), ('trip_end', 'Trip end'), ('rebalance_drop_off', 'Rebalance drop_off'), ('maintenance_drop_off', 'Maintenance drop off'), ('cancel_reservation', 'Cancel reservation'), ('reserve', 'Reserve'), ('trip_start', 'Trip start'), ('trip_enter', 'Trip enter'), ('trip_leave', 'Trip leave'), ('register', 'Register'), ('low_battery', 'Low battery'), ('maintenance', 'Maintenance'), ('service_end', 'Service end'), ('rebalance_pick_up', 'Rebalance pick up'), ('maintenance_pick_up', 'Maintenance pick up'), ('deregister', 'Deregister'), ('telemetry', 'Received telemetry'), ('battery_ok', 'Battery OK')]),
        ),
        migrations.AlterField(
            model_name='polygon',
            name='deletion_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='polygon',
            name='label',
            field=mds.models.UnboundedCharField(blank=True, null=True),
        ),
    ]



from rest_framework import serializers
from rest_framework import viewsets

from mds import models
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_PRV_API


class ProviderSerializer(serializers.ModelSerializer):
    id = serializers.UUIDField(
        required=True, help_text="Unique provider identifier (UUID)"
    )
    name = serializers.CharField(help_text="Name of the Provider")
    logo_b64 = serializers.CharField(
        required=False, help_text="Logo of provider base64 encoded"
    )

    class Meta:
        model = models.Provider
        fields = (
            "id",
            "name",
            "logo_b64",
            "base_api_url",
            "device_category",
            "api_authentication",
            "api_configuration",
            "agency_api_authentication",
        )


class ProviderViewSet(viewsets.ModelViewSet):
    permission_classes = (require_scopes(SCOPE_PRV_API),)
    queryset = models.Provider.objects.all()
    lookup_field = "id"
    serializer_class = ProviderSerializer



from django.contrib import admin
from . import models


@admin.register(models.Provider)
class ProviderAdmin(admin.ModelAdmin):
    list_display = ["id", "name"]
    ordering = ["name"]


@admin.register(models.Device)
class DeviceAdmin(admin.ModelAdmin):
    list_display = ["id", "provider", "identification_number", "model"]
    list_filter = ["provider", "model"]


@admin.register(models.EventRecord)
class EventRecordAdmin(admin.ModelAdmin):
    list_display = ["timestamp", "device", "event_type"]
    list_filter = ["device__provider", "event_type"]
    ordering = ["-timestamp"]


@admin.register(models.Area)
class AreaAdmin(admin.ModelAdmin):
    list_display = ["id", "creation_date", "deletion_date", "label"]
    list_filter = ["creation_date", "deletion_date"]
    ordering = ["-creation_date"]


@admin.register(models.Polygon)
class PolygonAdmin(admin.ModelAdmin):
    list_display = ["id", "creation_date", "deletion_date", "label"]
    list_filter = ["creation_date", "deletion_date"]
    ordering = ["-creation_date"]



import json

from rest_framework import serializers
from rest_framework import viewsets

from mds import enums
from mds import models
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_AGENCY_API
from mds.apis import utils


class MultiPolygonField(serializers.Field):
    def to_representation(self, value):
        ret = {"type": "MultiPolygon", "coordinates": []}
        for polygon in value:
            ret["coordinates"].append(json.loads(polygon.geom.geojson)["coordinates"])
        return ret

    def to_internal_value(self, value):
        raise NotImplementedError()


class AreaSerializer(serializers.ModelSerializer):
    """A service area
    """

    service_area_id = serializers.UUIDField(
        source="id", help_text="Unique Area identifier (UUID)"
    )
    start_date = utils.UnixTimestampMilliseconds(
        source="creation_date",
        help_text="Date at which this service area became effective",
    )
    end_date = utils.UnixTimestampMilliseconds(
        source="deletion_date",
        required=False,
        help_text="If exists, Date at which this service area was replaced.",
    )
    area = MultiPolygonField(source="polygons.all")
    type = serializers.ChoiceField(
        source="area_type",
        choices=enums.choices(enums.AREA_TYPE),
        default="unrestricted",
    )

    class Meta:
        model = models.Area
        fields = ("service_area_id", "start_date", "end_date", "area", "type")


class AreaViewSet(viewsets.ReadOnlyModelViewSet):
    permission_classes = (require_scopes(SCOPE_AGENCY_API),)
    queryset = models.Area.objects.prefetch_related("polygons").all()
    lookup_field = "id"
    serializer_class = AreaSerializer

    def get_queryset(self):
        queryset = super().get_queryset()
        provider_id = getattr(self.request.user, "provider_id", None)
        if provider_id:
            queryset = queryset.filter(providers__id=provider_id)
        else:
            queryset = queryset.none()

        return queryset



from django.utils.translation import ugettext as _

from rest_framework import exceptions
from rest_framework import serializers
from rest_framework.response import Response
from rest_framework.generics import GenericAPIView

from mds.apis import utils
from mds.access_control.permissions import require_scopes
from mds.access_control.scopes import SCOPE_PRV_API

import mds.authent.public_api as public_api


class AppCreationView(GenericAPIView):
    """ Implements an endpoint to create Oauth2 application for providers
    """

    permission_classes = (require_scopes(SCOPE_PRV_API),)

    class CreationRequestSerializer(serializers.Serializer):
        app_name = serializers.CharField()
        scopes = serializers.ListField(
            child=serializers.CharField(
                help_text="Scope of the application separated by a comma."
            )
        )
        app_owner = serializers.UUIDField(
            help_text="The owner for which the application is generated."
        )

    class CreationResponseSerializer(serializers.Serializer):
        client_id = serializers.CharField(help_text="Newly created Oauth app Client ID")
        client_secret = serializers.CharField(
            help_text="Newly created Oauth app Client Secret"
        )

    class RevocationRequestSerializer(serializers.Serializer):
        app_owner = serializers.UUIDField(
            help_text="The owner for which the application will be revoked."
        )
        delete = serializers.BooleanField(
            required=False,
            help_text="Indicate if the application should be removed or just revoked",
            default=False,
        )

    class RevokationResponseSerializer(utils.EmptyResponseSerializer):
        pass

    def post(self, request, *args, **kwargs):
        serializer = self.CreationRequestSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        validated_data = serializer.validated_data

        application = public_api.create_application(
            name=validated_data["app_name"],
            scopes=validated_data["scopes"],
            owner=validated_data["app_owner"],
        )

        serializer = self.CreationResponseSerializer(
            instance={
                "client_id": application["client_id"],
                "client_secret": application["client_secret"],
            }
        )

        return Response(serializer.data, status=200)

    def delete(self, request, *args, **kwargs):
        serializer = self.RevocationRequestSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        validated_data = serializer.validated_data

        try:
            if validated_data["delete"]:
                public_api.delete_application(validated_data["app_owner"])
            else:
                public_api.revoke_application(validated_data["app_owner"])
        except public_api.NoApplicationForOwner:
            raise exceptions.ValidationError(
                {
                    "app_owner": _("No application known for owner %s")
                    % validated_data["app_owner"]
                }
            )

        return Response(self.RevokationResponseSerializer().data, status=200)


class LongLivedTokenView(GenericAPIView):
    """Implements an endpoint to generate long lived (JWT) tokens.

    This is an alternate authentication method to Oauth2.
    An authorized user can generate such tokens and then transmit
    it to the token owner for future use with our API.
    """

    permission_classes = (require_scopes(SCOPE_PRV_API),)

    class RequestSerializer(serializers.Serializer):
        app_owner = serializers.UUIDField(
            help_text="The owner for which the token is generated."
        )
        token_duration = serializers.IntegerField(
            help_text="Token duration in seconds."
        )

    class RevokeRequestSerializer(serializers.Serializer):
        access_token = serializers.CharField()

    class RevokeResponseSerializer(utils.EmptyResponseSerializer):
        revocation_date = serializers.DateTimeField()

    class ResponseSerializer(serializers.Serializer):
        access_token = serializers.CharField()
        token_type = serializers.ChoiceField(choices=["bearer"])
        expires_in = serializers.IntegerField()

    def post(self, request, *args, **kwargs):
        serializer = self.RequestSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        validated_data = serializer.validated_data

        try:
            token = public_api.get_long_lived_token(
                validated_data["app_owner"], validated_data["token_duration"]
            )
        except public_api.NoApplicationForOwner:
            raise exceptions.ValidationError(
                {
                    "app_owner": _("No application known for owner %s")
                    % validated_data["app_owner"]
                }
            )
        serializer = self.ResponseSerializer(
            instance={
                "access_token": token,
                "token_type": "bearer",
                "expires_in": validated_data["token_duration"],
            }
        )
        return Response(serializer.data, status=200)

    def delete(self, request, *args, **kwargs):
        serializer = self.RevokeRequestSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        validated_data = serializer.validated_data

        try:
            revocation_date = public_api.revoke_long_lived_token(
                serializer.validated_data["access_token"]
            )
        except public_api.UnknownToken:
            raise exceptions.ValidationError(
                {
                    "token": _("No access token known for %s")
                    % validated_data["access_token"]
                }
            )

        return Response(
            self.RevokeResponseSerializer({"revocation_date": revocation_date}).data,
            status=200,
        )



# Generated by Django 2.1.4 on 2019-01-24 09:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authent', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accesstoken',
            name='revoked_after',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='application',
            name='owner',
            field=models.UUIDField(blank=True, db_index=True, help_text='Unique identifier for the owner of the application, not necessarily a django User.', null=True),
        ),
    ]



"""
Database description
"""
import datetime
import json
import uuid

from django import forms, utils
from django.contrib.gis.db import models as gis_models
from django.contrib.postgres import fields as pg_fields
from django.db import models
from django.db.models import OuterRef, Subquery, Prefetch
from django.utils import timezone

from rest_framework.utils import encoders

from . import enums


class UnboundedCharField(models.TextField):
    """Unlimited text, on a single line.

    Shows an ``<input type="text">`` in HTML but is stored as a TEXT
    column in Postgres (like ``TextField``).

    Like the standard :class:`~django.db.models.fields.CharField` widget,
    a ``select`` widget is automatically used if the field defines ``choices``.
    """

    def formfield(self, **kwargs):
        kwargs["widget"] = None if self.choices else forms.TextInput
        return super().formfield(**kwargs)


def agency_api_authentication_default():
    """Default value for the authentication of the Agency API use by the provider.

    Other keys are specific to each type:
    - token: "header" and "token"
    - oauth2: "client_id" and "client_secret"
    """
    return {"type": "none"}


def api_authentication_default():
    """Default value for the API authentication field of a provider.

    Other keys are specific to each type:
    - token: "header" and "token"
    - oauth2: "client_id" and "client_secret"
    """
    return {"type": "none"}


def api_configuration_default():
    """Default value for the API configuration field of a provider.

    Some provider implementations can be picky, configure these special cases here.
    """
    return {"trailing_slash": False}  # Some providers endpoint won't reply without it


def short_uuid4(uid):
    """When seeing a glimpse of the UID is enough."""
    return str(uid)[:8]  # Basically splitting on the first dash


class Provider(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    name = UnboundedCharField(default="")
    logo_b64 = UnboundedCharField(null=True, blank=True, default=None)
    device_category = UnboundedCharField(choices=enums.choices(enums.DEVICE_CATEGORY))

    # The following fields are for us pulling data from the provider
    base_api_url = UnboundedCharField(default="", verbose_name="Base Provider API URL")
    oauth2_url = UnboundedCharField(
        default="", verbose_name="OAuth2 URL (if different)"
    )
    api_authentication = pg_fields.JSONField(
        default=api_authentication_default, verbose_name="Provider API Authentication"
    )
    api_configuration = pg_fields.JSONField(
        default=api_configuration_default, verbose_name="Provider API Configuration"
    )
    # We may poll a provider, e.g. LADOT sandbox that replies for many providers
    # but has no device itself.
    # So we cannot just rely on checking the latest event record saved.
    last_start_time_polled = models.DateTimeField(blank=True, null=True)

    # The following fields are for the provider pushing data to us
    agency_api_authentication = pg_fields.JSONField(
        default=agency_api_authentication_default,
        verbose_name="Agency API Authentication",
    )

    def __str__(self):
        return "{} ({})".format(self.name, short_uuid4(self.id))


class DeviceQueryset(models.QuerySet):
    def with_latest_event(self):
        return self.prefetch_related(
            Prefetch(
                "event_records",
                queryset=EventRecord.objects.filter(
                    id__in=Subquery(
                        EventRecord.objects.filter(device_id=OuterRef("device_id"))
                        .exclude(event_type="telemetry")
                        .order_by("-timestamp")
                        .values_list("id", flat=True)[:1]
                    )
                ),
                to_attr="_latest_event",
            )
        )


class Device(models.Model):
    """A device
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    provider = models.ForeignKey(
        Provider, related_name="devices", on_delete=models.CASCADE
    )
    registration_date = models.DateTimeField(default=timezone.now)
    identification_number = UnboundedCharField()

    category = UnboundedCharField(choices=enums.choices(enums.DEVICE_CATEGORY))
    model = UnboundedCharField(default="")
    propulsion = pg_fields.ArrayField(
        UnboundedCharField(choices=enums.choices(enums.DEVICE_PROPULSION))
    )
    year_manufactured = models.IntegerField(null=True)
    manufacturer = UnboundedCharField(default="")

    # denormalized fields, the source of truth for this data is in the EventRecord table.
    # /!\ These fields are for internal usage and may disappear anytime
    dn_battery_pct = models.FloatField(null=True)
    dn_gps_point = gis_models.PointField(null=True)
    dn_gps_timestamp = models.DateTimeField(null=True)
    dn_status = UnboundedCharField(
        null=True, choices=enums.choices(enums.DEVICE_STATUS), default=None
    )

    objects = DeviceQueryset.as_manager()

    def __str__(self):
        return "{} {} ({})".format(
            self.get_category_display(),
            self.identification_number,
            short_uuid4(self.id),
        )

    @property
    def latest_event(self):
        if hasattr(self, "_latest_event"):
            # don't do a query in this case, the telemetry was prefetched.
            return self._latest_event[0] if self._latest_event else None
        device = Device.objects.filter(pk=self.pk).with_latest_event().get()
        return device.latest_event

    @property
    def gps_point_as_geojson(self):
        """Represent the GPS point as GeoJSON"""
        if not self.dn_gps_point:
            return None
        return json.loads(self.dn_gps_point.geojson)


class EventRecord(models.Model):
    timestamp = models.DateTimeField(db_index=True)
    point = gis_models.PointField(null=True)
    saved_at = models.DateTimeField(default=utils.timezone.now)
    source = models.CharField(
        choices=enums.choices(enums.EVENT_SOURCE),
        default=enums.EVENT_SOURCE.push.name,
        max_length=16,
    )
    device = models.ForeignKey(
        Device, related_name="event_records", on_delete=models.CASCADE
    )
    event_type = UnboundedCharField(choices=enums.choices(enums.EVENT_TYPE))

    # JSON fields:
    # {
    #   "telemetry": see factories.EventRecord for an exhaustive example
    # }
    properties = pg_fields.JSONField(default=dict, encoder=encoders.JSONEncoder)

    class Meta:
        unique_together = [("device", "timestamp")]

    @property
    def point_as_geojson(self):
        """Represent the GPS point as GeoJSON"""
        if not self.point:
            return None
        return json.loads(self.point.geojson)

    @property
    def updated_status(self):
        return enums.EVENT_TYPE_TO_DEVICE_STATUS.get(
            self.event_type, enums.DEVICE_STATUS.unknown.name
        )


class Polygon(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    creation_date = models.DateTimeField(default=utils.timezone.now)
    deletion_date = models.DateTimeField(null=True, blank=True)
    label = UnboundedCharField(default="", blank=True, db_index=True)
    geom = gis_models.PolygonField()
    properties = pg_fields.JSONField(default=dict, encoder=encoders.JSONEncoder)

    def __str__(self):
        return "{} ({})".format(self.label, short_uuid4(self.id))


class Area(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    creation_date = models.DateTimeField(
        default=datetime.datetime(2012, 1, 1, tzinfo=datetime.timezone.utc)
    )
    deletion_date = models.DateTimeField(null=True, blank=True)
    label = UnboundedCharField(default="", blank=True, db_index=True)
    polygons = models.ManyToManyField(Polygon, blank=True, related_name="areas")
    providers = models.ManyToManyField(Provider, blank=True, related_name="areas")
    color = models.CharField(
        max_length=8, default="#FFFFFF", help_text="hexa representation"
    )
    area_type = UnboundedCharField(
        choices=enums.choices(enums.AREA_TYPE), default="unrestricted"
    )

    def __str__(self):
        return "{} ({})".format(self.label, short_uuid4(self.id))



from django.conf.urls import include
from django.contrib import admin
from django.contrib.auth import views as auth_views
from django.urls import path

import mds.apis.agency_api.v0_x.urls
import mds.apis.prv_api.urls
import mds.authent.urls

urlpatterns = [
    path(
        "mds/v0.x/",
        include(
            (
                mds.apis.agency_api.v0_x.urls.get_url_patterns(prefix="mds/v0.x"),
                "agency",
            )
        ),
    ),
    path("prv/", include(mds.apis.prv_api.urls, namespace="private")),
    path("admin/", admin.site.urls),
    path("authent/", include(mds.authent.urls, namespace="authent")),
    # oauth2_provider gives views to manage applications.
    # These views require a logged in user.
    path(
        "accounts/login/",
        auth_views.LoginView.as_view(template_name="admin/login.html"),
    ),
]



# Generated by Django 2.1.3 on 2019-01-22 09:25

import django.contrib.gis.db.models.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0003_record_timestamp'),
    ]

    operations = [
        migrations.AddField(
            model_name='eventrecord',
            name='point',
            field=django.contrib.gis.db.models.fields.PointField(null=True, srid=4326),
        ),
    ]



"""
Django settings
"""
import itertools
import os

import getconf
from corsheaders.defaults import default_headers

from mds.access_control.auth_means import (
    SecretKeyJwtBaseAuthMean,
    PublicKeyJwtBaseAuthMean,
)

CONFIG = getconf.ConfigGetter("mds")

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = CONFIG.getstr(
    "django.secret_key", "weak", doc="Secret key for signing data"
)
DEBUG = CONFIG.getbool("dev.debug", False)

ALLOWED_HOSTS = CONFIG.getlist(
    "django.allowed_hosts", [], doc="Comma-separated list of allowed hosts"
)

INSTALLED_APPS = [
    "mds.authent.apps.Config",
    "mds.authent.oauth2_provider_apps.Config",
    "mds.apps.Config",
    "corsheaders",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.postgres",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "rest_framework",
    "django_filters",
    "drf_yasg",
]
MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "corsheaders.middleware.CorsMiddleware",
]
ROOT_URLCONF = "mds.urls"
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]
WSGI_APPLICATION = "mds.server.wsgi.application"
DATABASES = {
    "default": {
        "ENGINE": "django.contrib.gis.db.backends.postgis",
        "NAME": CONFIG.getstr("db.name", "mds"),
        "USER": CONFIG.getstr("db.user", "postgres"),
        "PASSWORD": CONFIG.getstr("db.password"),
        "HOST": CONFIG.getstr("db.host", "127.0.0.1"),
        "PORT": CONFIG.getstr("db.port", "5432"),
        "ATOMIC_REQUESTS": True,
    }
}

# Password validation
AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"}
]
LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_L10N = True
USE_TZ = True
STATIC_URL = "/static/"
LOCALE_PATH = (os.path.join(BASE_DIR, "locale"),)

CORS_ORIGIN_ALLOW_ALL = True
CORS_ALLOW_CREDENTIALS = True
CORS_ALLOW_HEADERS = list(default_headers) + ["cache-control"]

REST_FRAMEWORK = {
    # "DEFAULT_PERMISSION_CLASSES": ("rest_framework.permissions.IsAuthenticated",),
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "mds.access_control.stateless_jwt.StatelessJwtAuthentication",
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.BasicAuthentication",
    )
}
SWAGGER_SETTINGS = {
    "DEFAULT_AUTO_SCHEMA_CLASS": "mds.apis.schema.CustomSwaggerAutoSchema",
    "SECURITY_DEFINITIONS": {
        "Bearer": {"type": "apiKey", "name": "Authorization", "in": "header"}
    },
}

# AUTH_USER_MODEL = None # TODO later

AUTH_MEANS = []
for i in itertools.count(start=1):
    section = "auth" + ("_" + str(i) if i > 1 else "")

    auth_mean = None
    if CONFIG.getstr(section + ".secret_key"):
        auth_mean = SecretKeyJwtBaseAuthMean(CONFIG.getstr(section + ".secret_key"))
    elif CONFIG.getstr(section + ".public_key"):
        # PEM public keys (-----BEGIN PUBLIC KEY-----)
        auth_mean = PublicKeyJwtBaseAuthMean(CONFIG.getstr(section + ".public_key"))
    else:
        break

    # Optional, recommended if handling long-lived access tokens
    # (not a good idea when using JWT)
    if "introspect_url" in section:
        auth_mean.introspect_url = CONFIG.getstr(section + ".introspect_url", None)

    AUTH_MEANS.append(auth_mean)
AUTHENT_SECRET_KEY = "my-secret"
AUTHENT_RSA_PRIVATE_KEY = ""
OAUTH2_PROVIDER = {
    "APPLICATION_MODEL": "authent.Application",
    "ACCESS_TOKEN_MODEL": "authent.AccessToken",
    "GRANT_MODEL": "authent.Grant",
    "REFRESH_TOKEN_MODEL": "authent.RefreshToken",
    "OAUTH2_SERVER_CLASS": "mds.authent.oauthlib_utils.Server",
    "OAUTH2_VALIDATOR_CLASS": "mds.authent.oauthlib_utils.OAuth2Validator",
    "SCOPES_BACKEND_CLASS": "mds.authent.oauthlib_utils.AppScopes",
}
MIGRATION_MODULES = {
    # see https://github.com/jazzband/django-oauth-toolkit/issues/634
    # swappable models are badly designed in oauth2_provider
    # ignore migrations and provide our own models.
    "oauth2_provider": None
}



# Generated by Django 2.1.3 on 2019-01-25 10:14

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0005_auto_20190122_1300'),
    ]

    operations = [
        migrations.AddField(
            model_name='provider',
            name='device_category',
            field=mds.models.UnboundedCharField(choices=[('bike', 'Bike'), ('scooter', 'Scooter'), ('car', 'Car')], default='scooter'),
            preserve_default=False,
        ),
    ]



from django.contrib.postgres import fields as pg_fields
from django.db import models
from oauth2_provider.models import (
    AbstractApplication,
    AbstractAccessToken,
    AbstractGrant,
    AbstractRefreshToken,
)


class Application(AbstractApplication):
    owner = models.UUIDField(
        null=True,
        blank=True,
        db_index=True,
        help_text="Unique identifier for the owner of the application, not necessarily a django User.",
    )
    scopes = pg_fields.ArrayField(
        models.CharField(max_length=32), help_text="Application allowed scopes."
    )

    def natural_key(self):
        return (self.client_id,)

    @property
    def scopes_string(self):
        return " ".join(sorted(set(self.scopes)))


class AccessToken(AbstractAccessToken):
    token = models.TextField()  # remove 255 char limit (for JWT)
    jti = models.UUIDField(db_index=True)
    revoked_after = models.DateTimeField(null=True, blank=True)


class Grant(AbstractGrant):
    pass


class RefreshToken(AbstractRefreshToken):
    pass



# Generated by Django 2.1.7 on 2019-03-01 11:13

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0010_device_dn_battery_pct'),
    ]

    operations = [
        migrations.AlterField(
            model_name='area',
            name='label',
            field=mds.models.UnboundedCharField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='polygon',
            name='label',
            field=mds.models.UnboundedCharField(blank=True, db_index=True, null=True),
        ),
    ]



# Generated by Django 2.1.4 on 2019-01-17 00:08

import django.contrib.gis.db.models.fields
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import mds.models
import rest_framework.utils.encoders
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Area",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                (
                    "creation_date",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("deletion_date", models.DateTimeField(null=True)),
                ("label", mds.models.UnboundedCharField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Device",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                (
                    "registration_date",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("identification_number", mds.models.UnboundedCharField()),
                (
                    "category",
                    mds.models.UnboundedCharField(
                        choices=[
                            ("bike", "Bike"),
                            ("scooter", "Scooter"),
                            ("car", "Car"),
                        ]
                    ),
                ),
                ("model", mds.models.UnboundedCharField(default=str)),
                (
                    "propulsion",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=mds.models.UnboundedCharField(
                            choices=[
                                ("electric", "Electric"),
                                ("combustion", "Combustion"),
                            ]
                        ),
                        size=None,
                    ),
                ),
                ("year_manufactured", models.IntegerField(null=True)),
                ("manufacturer", mds.models.UnboundedCharField(default=str)),
                (
                    "dn_gps_point",
                    django.contrib.gis.db.models.fields.PointField(
                        null=True, srid=4326
                    ),
                ),
                ("dn_gps_timestamp", models.DateTimeField(null=True)),
                (
                    "dn_status",
                    mds.models.UnboundedCharField(
                        choices=[
                            ("available", "Available"),
                            ("reserved", "Reserved"),
                            ("unavailable", "Unavailable"),
                            ("removed", "Removed"),
                        ],
                        default="unavailable",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EventRecord",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "saved_at",
                    models.DateTimeField(
                        db_index=True, default=django.utils.timezone.now
                    ),
                ),
                (
                    "source",
                    models.CharField(
                        choices=[("push", "push"), ("pull", "pull")],
                        default="push",
                        max_length=16,
                    ),
                ),
                (
                    "event_type",
                    mds.models.UnboundedCharField(
                        choices=[
                            ("service_start", "Service start"),
                            ("trip_end", "Trip end"),
                            ("rebalance_drop_off", "Rebalance drop_off"),
                            ("maintenance_drop_off", "Maintenance drop off"),
                            ("cancel_reservation", "Cancel reservation"),
                            ("reserve", "Reserve"),
                            ("trip_start", "Trip start"),
                            ("trip_enter", "Trip enter"),
                            ("trip_leave", "Trip leave"),
                            ("register", "Register"),
                            ("low_battery", "Low battery"),
                            ("maintenance", "Maintenance"),
                            ("service_end", "Service end"),
                            ("rebalance_pick_up", "Rebalance pick up"),
                            ("maintenance_pick_up", "Maintenance pick up"),
                            ("deregister", "Deregister"),
                            ("telemetry", "Received telemetry"),
                        ]
                    ),
                ),
                (
                    "properties",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        default=dict, encoder=rest_framework.utils.encoders.JSONEncoder
                    ),
                ),
                (
                    "device",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="event_records",
                        to="mds.Device",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Polygon",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                (
                    "creation_date",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("deletion_date", models.DateTimeField(null=True)),
                ("label", mds.models.UnboundedCharField(null=True)),
                ("geom", django.contrib.gis.db.models.fields.PolygonField(srid=4326)),
                (
                    "properties",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        default=dict, encoder=rest_framework.utils.encoders.JSONEncoder
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Provider",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("name", mds.models.UnboundedCharField(default=str)),
                (
                    "logo_b64",
                    mds.models.UnboundedCharField(blank=True, default=None, null=True),
                ),
            ],
        ),
        migrations.AddField(
            model_name="device",
            name="provider",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="devices",
                to="mds.Provider",
            ),
        ),
        migrations.AddField(
            model_name="area",
            name="polygons",
            field=models.ManyToManyField(
                blank=True, related_name="areas", to="mds.Polygon"
            ),
        ),
        migrations.AddField(
            model_name="area",
            name="providers",
            field=models.ManyToManyField(
                blank=True, related_name="areas", to="mds.Provider"
            ),
        ),
    ]



import datetime


def to_mds_timestamp(dt: datetime.datetime) -> int:
    """Convert a datetime object into an MDS-compliant timestamp.

    MDS wants the timestamp to be in milliseconds.
    """
    return round(dt.timestamp() * 1000)


def from_mds_timestamp(value: int) -> datetime:
    """Convert an MDS timestamp (in milliseconds) to a datetime object."""
    return datetime.datetime.fromtimestamp(value / 1000, tz=datetime.timezone.utc)



# Generated by Django 2.1.5 on 2019-03-19 09:49

from django.db import migrations
import mds.models


class Migration(migrations.Migration):

    dependencies = [
        ('mds', '0014_non_null_labels'),
    ]

    operations = [
        migrations.AlterField(
            model_name='device',
            name='dn_status',
            field=mds.models.UnboundedCharField(choices=[('available', 'Available'), ('reserved', 'Reserved'), ('unavailable', 'Unavailable'), ('removed', 'Removed'), ('trip', 'Trip'), ('elsewhere', 'Elsewhere'), ('inactive', 'Inactive'), ('unknown', 'Unknown')], default=None, null=True),
        ),
    ]



from django.conf import settings
from django.utils.encoding import smart_text
from rest_framework.authentication import BaseAuthentication, get_authorization_header

from .authenticate import authenticate


class StatelessJwtAuthentication(BaseAuthentication):
    def authenticate_header(self, request):
        return "Bearer"

    def authenticate(self, request):
        encoded_jwt = self.extract_token(request)
        if encoded_jwt is None:
            return None

        if not settings.AUTH_MEANS:
            raise Exception(
                "JWT authentication configuration is incomplete: "
                + "neither secret nor public key found"
            )

        user = authenticate(settings.AUTH_MEANS, encoded_jwt)

        return user, None

    @staticmethod
    def extract_token(request):
        auth_header = get_authorization_header(request)

        if not auth_header:
            return None

        auth_header_prefix = "Bearer ".lower()
        auth = smart_text(auth_header)

        if not auth.lower().startswith(auth_header_prefix):
            return None

        return auth[len(auth_header_prefix) :]



"""
Pulling data for registered providers

This is the opposite of provider pushing their data to the agency API.
"""
import json
import logging
import threading
import traceback
import urllib.parse
import uuid

from django.contrib.gis import geos
from django.core import management
from django.db import connection
from django.db import transaction
from django.utils import timezone

from oauthlib.oauth2 import BackendApplicationClient
import requests
from requests_oauthlib import OAuth2Session
from retrying import retry
from semantic_version import Version

from mds import enums
from mds import models
from mds import utils


ACCEPTED_MDS_VERSIONS = ["0.2", "0.3"]

logger = logging.getLogger(__name__)


class OAuth2Store(threading.local):
    def __init__(self):
        super().__init__()
        self.token_cache = {}

    def get_client(self, provider):
        """Authenticate using the Backend Application Flow from OAuth2."""
        client_id = provider.api_authentication["client_id"]
        token = self._get_access_token(provider)
        client = OAuth2Session(client_id, token=token)
        return client

    def _get_access_token(self, provider):
        """Fetch and cache the token."""
        if provider not in self.token_cache:
            self.token_cache[provider] = self._fetch_access_token(provider)
        return self.token_cache[provider]

    @retry(stop_max_attempt_number=2)
    def _fetch_access_token(self, provider):
        """Fetch an access token from the provider"""
        client_id = provider.api_authentication["client_id"]
        client_secret = provider.api_authentication["client_secret"]
        client = BackendApplicationClient(client_id=client_id)
        session = OAuth2Session(client=client)

        # The URL to fetch the token maybe on another base URL
        token_url = provider.oauth2_url
        if not token_url:
            token_url = urllib.parse.urljoin(provider.base_api_url, "/oauth2/token")
            if provider.api_configuration.get("trailing_slash"):
                token_url += "/"

        # Provider-specific parameters to request the token
        kwargs = {}
        try:
            kwargs.update(provider.api_configuration["token_params"])
        except KeyError:
            pass

        token = session.fetch_token(
            token_url=token_url,
            client_id=client_id,
            client_secret=client_secret,
            **kwargs
        )
        return token


oauth2_store = OAuth2Store()


class Command(management.BaseCommand):
    help = "Poll providers to fetch their latest data."

    def handle(self, *args, **options):
        self.verbosity = options["verbosity"]
        self.token_cache = {}
        self.providers = set(models.Provider.objects.values_list("pk", flat=True))
        self.devices = set(models.Device.objects.values_list("pk", flat=True))

        for provider in models.Provider.objects.all():
            if not provider.base_api_url:
                self.stdout.write(
                    "Provider %s has no URL defined, skipping." % provider.name
                )
                continue

            self.stdout.write("Polling %s... " % provider.name, ending="")

            try:
                self.poll_status_changes(provider)
            except Exception:  # pylint: disable=broad-except
                logger.exception("Error in polling provider %s", provider.name)
                self.stderr.write("Polling failed!")
                self.stderr.write(traceback.format_exc())
                # Try the next provider anyway
                continue

            self.stdout.write("Success!")

    def poll_status_changes(self, provider):
        next_url = urllib.parse.urljoin(provider.base_api_url, "status_changes")
        if provider.api_configuration.get("trailing_slash"):
            next_url += "/"

        params = {}

        # Start where we left, it's all based on providers sorting by start_time, obviously
        # (we wouldn't know if the provider recorded late telemetries earlier than this date)
        if provider.last_start_time_polled:
            params["start_time"] = utils.to_mds_timestamp(
                provider.last_start_time_polled
            )

        # Provider-specific params to optimise polling
        try:
            params.update(provider.api_configuration["status_changes_params"])
        except KeyError:
            pass

        if params:
            next_url = "%s?%s" % (next_url, urllib.parse.urlencode(params))

        # Pagination
        while next_url:
            body = self.get_body(provider, next_url)
            # Translate older versions of data
            translated_data = translate_data(body["data"], body["version"])
            status_changes = translated_data["status_changes"]
            if not status_changes:
                break

            # A transaction for each "page" of data
            with transaction.atomic():
                last_start_time_polled = self.process_status_changes(
                    status_changes, provider
                )
                provider.last_start_time_polled = last_start_time_polled
                provider.save()

            next_url = body.get("links", {}).get("next")

    @retry(stop_max_attempt_number=2)
    def get_body(self, provider, url):
        authentication_type = provider.api_authentication.get("type")
        if authentication_type in (None, "", "none"):
            client = requests
        elif authentication_type == "oauth2":
            client = oauth2_store.get_client(provider)
        else:
            raise NotImplementedError

        headers = {}
        accepted_content_types = ["application/json"]
        # Add versioning header (some providers may choke on this)
        for version in ACCEPTED_MDS_VERSIONS:
            accepted_content_types.append(
                "application/vnd.mds.provider+json;version=%s" % version
            )
        headers["Accept"] = ", ".join(accepted_content_types)

        if self.verbosity > 1:
            self.stdout.write("Polling provider on URL %s" % url)
        response = client.get(url, timeout=30, headers=headers)
        # When we get a 401 with oauth2, we should try renewing the access token,
        # but when we try every minute, we're just delaying that retrial
        response.raise_for_status()

        # Servers should send what version of the API was served
        # but we rather trust the versioning in the body
        content_type = response.headers.get(
            "Content-Type", "application/json; charset=UTF-8"
        )
        if content_type.startswith("application/json"):
            version = "unspecified"
        else:
            # Let it raise if malformed
            content_type, version = content_type.split(";")
            _, version = version.split("=")
        if self.verbosity > 1:
            self.stdout.write("Provider pretends to accept version %s" % version)

        body = response.json()
        return body

    def process_status_changes(self, status_changes, provider):
        if self.verbosity > 1:
            self.stdout.write("Processing...")
        self.prepare_status_changes(status_changes, provider)
        self.create_missing_providers(status_changes)
        self.create_missing_devices(status_changes)
        self.create_event_records(status_changes)

        # Returning the latest event recorded (order is not asserted in the specs)
        last_start_time_polled = utils.from_mds_timestamp(
            max(status_change["event_time"] for status_change in status_changes)
        )
        return last_start_time_polled

    def prepare_status_changes(self, status_changes, provider):
        """Some preliminary checks/addenda"""

        for status_change in status_changes:
            status_change["provider_id"] = uuid.UUID(status_change["provider_id"])
            status_change["device_id"] = device_id = uuid.UUID(
                status_change["device_id"]
            )
            status_change["event_time"] = int(status_change["event_time"])

            # The list of event types and even the naming don't match between
            # the provider and agency APIs, so translate one to the other
            try:
                event_type_reason = status_change["event_type_reason"]
            except KeyError:  # Spec violation!
                self.stderr.write(
                    "Warning: device %s has no event_type_reason" % device_id
                )
                # Let recording the device fail with a null value, but only that one in the batch
                agency_event_type = None
            else:
                agency_event_type = enums.PROVIDER_EVENT_TYPE_REASON_TO_AGENCY_EVENT_TYPE[
                    event_type_reason
                ]
            status_change["agency_event_type"] = agency_event_type

            event_location = status_change["event_location"]
            if event_location:
                assert event_location["geometry"]["type"] == "Point"
                # Some providers may get the (lng, lat) order wrong
                if provider.api_configuration.get("swap_lng_lat"):
                    lat, lng = event_location["geometry"]["coordinates"][:2]
                    event_location["geometry"]["coordinates"][0] = lng
                    event_location["geometry"]["coordinates"][1] = lat
            else:  # Spec violation!
                self.stderr.write(
                    "Warning: device %s has no event_location" % device_id
                )

    def create_missing_providers(self, status_changes):
        """Make sure all providers mentioned exist"""

        for status_change in status_changes:
            provider_id = status_change["provider_id"]
            if provider_id not in self.providers:
                name = status_change["provider_name"]
                device_category = status_change["vehicle_type"]

                models.Provider.objects.create(
                    pk=provider_id, name=name, device_category=device_category
                )
                self.providers.add(provider_id)
                self.stdout.write("Provider %s was created." % name)

    def create_missing_devices(self, status_changes):
        """Make sure all devices mentioned exist"""

        with_missing_devices = [
            status_change
            for status_change in status_changes
            if status_change["device_id"] not in self.devices
        ]

        with connection.cursor() as cursor:
            cursor.executemany(
                """INSERT INTO mds_device (
                    id,
                    provider_id,
                    registration_date,
                    identification_number,
                    category,
                    model,
                    propulsion,
                    manufacturer
                ) VALUES (
                    %(id)s,
                    %(provider_id)s,
                    %(registration_date)s,
                    %(identification_number)s,
                    %(category)s,
                    %(model)s,
                    %(propulsion)s,
                    %(manufacturer)s
                ) ON CONFLICT DO NOTHING
                """,
                (
                    create_device(status_change)
                    for status_change in with_missing_devices
                ),
            )

        devices_added = [
            str(status_change["device_id"]) for status_change in with_missing_devices
        ]
        if devices_added:
            self.devices.update(devices_added)
            self.stdout.write("Devices created: %s" % ", ".join(devices_added))

    def create_event_records(self, status_changes):
        """Now record the... records"""

        # Timestamps are unique per device, ignore duplicates
        with connection.cursor() as cursor:
            # Using "upsert"
            cursor.executemany(
                """INSERT INTO mds_eventrecord (
                    device_id,
                    timestamp,
                    source,
                    point,
                    event_type,
                    properties,
                    saved_at
                ) VALUES (
                    %(device_id)s,
                    %(timestamp)s,
                    %(source)s,
                    %(point)s,
                    %(event_type)s,
                    %(properties)s,
                    %(saved_at)s
                ) ON CONFLICT DO NOTHING
                """,
                (
                    create_event_record(status_change)
                    for status_change in status_changes
                ),
            )


def create_device(status_change):
    device_id = status_change["device_id"]
    identification_number = status_change["vehicle_id"]
    if not identification_number:  # Spec violation!
        identification_number = "test-%s" % str(device_id).split("-", 1)

    return {
        "id": device_id,
        # Don't assume the device received belongs to the provider requested
        # The LA sandbox contains data for several providers
        "provider_id": status_change["provider_id"],
        "registration_date": timezone.now(),
        "identification_number": identification_number,
        "category": status_change["vehicle_type"],
        "model": "",
        "propulsion": status_change["propulsion_type"],
        "manufacturer": "",
    }


def create_event_record(status_change):
    properties = {"trip_id": status_change.get("associated_trip")}

    event_location = status_change["event_location"]
    if event_location:
        # GeoJSON Point Feature
        try:
            longitude, latitude, altitude = event_location["geometry"]["coordinates"]
        except ValueError:
            longitude, latitude = event_location["geometry"]["coordinates"]
            altitude = None
        point = geos.Point(longitude, latitude, altitude, srid=4326)
        properties["telemetry"] = {
            "timestamp": event_location["properties"]["timestamp"],
            "gps": {"lng": longitude, "lat": latitude},
            # No coordinates, no battery charge saved
            "battery_pct": status_change.get("battery_pct"),
        }
        if altitude:
            properties["telemetry"]["gps"]["altitude"] = altitude
    else:  # Spec violation!
        point = None

    return {
        "device_id": status_change["device_id"],
        "timestamp": utils.from_mds_timestamp(status_change["event_time"]),
        "source": "pull",  # pulled by agency,
        "point": point.ewkt if point else None,
        "event_type": status_change["agency_event_type"],
        "properties": json.dumps(properties),
        "saved_at": timezone.now(),
    }


def translate_data(data, version):
    version = Version(version)

    if version >= Version("0.3.0") and version < Version("0.4.0"):
        return data

    if version >= Version("0.2.0"):
        # The only two noticeable changes from our point of view are:
        # - timestamps converted from floating-point seconds to milliseconds;
        # - "trip_ids" now is a single "trip_id"
        for status_change in data["status_changes"]:
            # We were already expecting milliseconds in the 0.2 implementation
            if "." in str(status_change["event_time"]):
                status_change["event_time"] = round(status_change["event_time"] * 1000)
            # Keep only the first (and probably only) trip ID
            associated_trip = status_change.pop("associated_trips", None)
            if associated_trip:
                status_change["associated_trip"] = associated_trip[0]
        return data

    raise NotImplementedError(version)
